/*
 * Test directory reading operations, exploring concurrency issues.
 *
 * Copyright (C) 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a directory name.\n");
    return 1;
  }

  char *filename = argv[1];

  printf("Opening %s...\n", filename);

  file_t *reader = client_opendir(filename);

  if (reader == NULL)
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  printf("Reading...\n");

  struct dirent *dirent;
  int filenum = 1000, remaining = 100;
  char buffer[strlen(filename) + strlen("/file-XXXX.txt") + 10];
  char data[256];
  int found = 0;
  file_t *files[100];

  while ((dirent = client_readdir(reader)) != NULL)
  {
    free(dirent);
    found++;

    /* Create files while getting the listing. */

    if (remaining)
    {
      sprintf(buffer, "%s/file-%d.txt", filename, filenum++);

      file_t *file = client_open(buffer, O_RDWR | O_CREAT);

      if (file == NULL)
      {
        printf("Could not open file: %s\n", buffer);
        return 1;
      }

      /* Write something to each file. */

      sprintf(data, "Data in file-%d.txt", filenum - 1);
      client_write(file, data, strlen(data));

      /* Remember the file for later. */

      files[remaining - 1] = file;

      remaining--;
    }
  }

  printf("Files found: %d\n", found);

  /* Re-read, counting files. */

  reader = client_opendir(filename);

  if (reader == NULL)
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  found = 0;

  while ((dirent = client_readdir(reader)) != NULL)
  {
    free(dirent);
    found++;
  }

  printf("Files found: %d\n", found);

  /* Remove some files. */

  for (filenum = 1000, remaining = 100; remaining > 0; remaining--)
  {
    sprintf(buffer, "%s/file-%d.txt", filename, filenum++);

    long err = client_remove(buffer);

    if (err)
    {
      printf("Could not remove file: %s\n", buffer);
      return 1;
    }
  }

  /* Re-read, counting files. */

  reader = client_opendir(filename);

  if (reader == NULL)
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  found = 0;

  while ((dirent = client_readdir(reader)) != NULL)
  {
    free(dirent);
    found++;
  }

  printf("Files found: %d\n", found);

  /* Read from the still-open but now removed files. */

  for (remaining = 0; remaining < 100; remaining++)
  {
    file_t *file = files[remaining];

    client_seek(file, SEEK_SET, 0);
    offset_t nread = client_read(file, data, 256);

    data[nread] = '\0';
    printf("Read: %s\n", data);
  }

  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
